#include <QtCore>
#include <QtGui>

#include <cstring>

#include <sane/sane.h>
#include <sane/saneopts.h>

const char *u = 0;
const char *p = 0;

void auth(SANE_String_Const resource, SANE_Char username[SANE_MAX_USERNAME_LEN], SANE_Char password[SANE_MAX_PASSWORD_LEN])
{
    qDebug("auth against %s", resource);

    strcpy(username, u);
    strcpy(password, p);
}

int main(int argc, char *argv[])
{
    QGuiApplication app(argc, argv);

    u = argv[1];
    p = argv[2];

    SANE_Int version;
    SANE_Status status;;

    if((status = sane_init(&version, auth)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_init() returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    qDebug("initialized version %d.%d.%d",
        SANE_VERSION_MAJOR(version),
        SANE_VERSION_MINOR(version),
        SANE_VERSION_BUILD(version));

    const SANE_Device **odevices, **devices, *device;

    if((status = sane_get_devices(&odevices, SANE_FALSE)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_get_devices() returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    int index = 0;
    devices = odevices;

    if(!*devices)
    {
        qDebug("fatal: no scanners found, exiting");
        return 1;
    }

    int num = 0;

    while((device = *devices++))
    {
        num++;
        qDebug("device #%d '%s' | '%s' | '%s' | '%s'", index++, device->name, device->vendor, device->model, device->type);
    }

    if(num > 1)
    {
        do
        {
            qDebug("Index to use:");
        } while(scanf("%d", &index) < 1);
    }
    else
        index = 0;

    SANE_Handle handle;

    if((status = sane_open(odevices[index]->name, &handle)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_open() returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    SANE_Int res;
    const SANE_Option_Descriptor *desc;
    index = 0;

    typedef QHash<QString, int> Options;
    Options options;

    while((desc = sane_get_option_descriptor(handle, index)))
    {
        qDebug("option '%s' | '%s' | '%s'", desc->name, desc->title, desc->desc);

        options.insert(desc->name, index);
        index++;
    }

    int v = 300;
    SANE_Char data[32];

    if((status = sane_control_option(handle, options[SANE_NAME_SCAN_RESOLUTION], SANE_ACTION_SET_VALUE, &v, &res)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_control_option() returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    if((status = sane_control_option(handle, options[SANE_NAME_SCAN_RESOLUTION], SANE_ACTION_GET_VALUE, &v, &res)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_control_option() returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    qDebug("DPI %d", v);

    // set color mode
    if((status = sane_control_option(handle, options[SANE_NAME_SCAN_MODE], SANE_ACTION_SET_VALUE, const_cast<char *>(SANE_VALUE_SCAN_MODE_COLOR), &res)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_control_option() [set color] returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    const char *mode = "Duplex"; // ADF / Duplex

    // set source
    // Automatic Document Feeder
    if((status = sane_control_option(handle, options[SANE_NAME_SCAN_SOURCE], SANE_ACTION_SET_VALUE, const_cast<char *>(mode), &res)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_control_option() [set source] returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    // get source
    if((status = sane_control_option(handle, options[SANE_NAME_SCAN_SOURCE], SANE_ACTION_GET_VALUE, data, &res)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_control_option() [get source] returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    qDebug("Scan source: %s", data);

    // image format
    if((status = sane_control_option(handle, options[SANE_NAME_SCAN_MODE], SANE_ACTION_GET_VALUE, data, &res)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_control_option() returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    QImage::Format format = QImage::Format_Invalid;
    QVector<QRgb> table;

    qDebug("Scan mode: '%s'", data);

    /*
     * SANE_VALUE_SCAN_MODE_COLOR
     * SANE_VALUE_SCAN_MODE_COLOR_LINEART
     * SANE_VALUE_SCAN_MODE_COLOR_HALFTONE
     * SANE_VALUE_SCAN_MODE_GRAY
     * SANE_VALUE_SCAN_MODE_HALFTONE
     * SANE_VALUE_SCAN_MODE_LINEART
     */
    if(!strcmp(data, SANE_VALUE_SCAN_MODE_COLOR) || !strcmp(data, "Full Color"))
        format = QImage::Format_RGB888;
    else if(!strcmp(data, SANE_VALUE_SCAN_MODE_GRAY))
    {
        format = QImage::Format_Indexed8;

        for(int i = 0;i < 256;i++)
            table.append(qRgb(i, i, i));
    }
    else
    {
        qDebug("fatal: unsupported scan mode '%s'", data);
        return 1;
    }

    int id = 0;

    while(true)
    {

    SANE_Byte buffer[8192];
    SANE_Int stored;
    // start scanning
    if((status = sane_start(handle)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_start() returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    QBuffer mem;
    int total = 0;

    if(!mem.open(QIODevice::ReadWrite | QIODevice::Truncate))
    {
        qDebug("fatal: cannot open file (%s)", qPrintable(mem.errorString()));
        return 1;
    }

    do
    {   
        stored = 0;
        status = sane_read(handle, buffer, sizeof(buffer), &stored);

        total += stored;

        if(stored && mem.write(reinterpret_cast<const char *>(buffer), stored) != stored)
        {
            qWarning("Cannot write data into the buffer (%s)", qPrintable(mem.errorString()));
            sane_cancel(handle);
            return 1;
        }
    } while(status == SANE_STATUS_GOOD);

    if(status == SANE_STATUS_EOF)
        qDebug("sane_read() finished OK with status %d (%s)", status, sane_strstatus(status));
    else
    {
        qDebug("fatal: sane_read() failed with status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    mem.seek(0);

    qDebug("Saved %d/%d bytes in the buffer", total, mem.data().length());

    {
        QFile f("scanned-data.bin");
        f.open(QIODevice::ReadWrite | QIODevice::Truncate);
        f.write(mem.buffer());
        f.close();
    }

    // scanner params
    SANE_Parameters params;

    if((status = sane_get_parameters(handle, &params)) != SANE_STATUS_GOOD)
    {
        qDebug("fatal: sane_get_parameters() returned status %d (%s)", status, sane_strstatus(status));
        return 1;
    }

    qDebug("format %d", params.format);
    qDebug("last_frame %s", params.last_frame ? "yes" : "no");
    qDebug("bytes_per_line %d", params.bytes_per_line);
    qDebug("pixels_per_line %d", params.pixels_per_line);
    qDebug("lines %d", params.lines);
    qDebug("depth %d", params.depth);

    if(!params.pixels_per_line || !params.bytes_per_line)
    {
        qDebug("fatal: invalid image parameters");
        return 1;
    }

    QImage image(reinterpret_cast<const uchar *>(mem.data().constData()),
                    params.pixels_per_line,
                    params.lines,
                    params.bytes_per_line,
                    format);

    if(!table.isEmpty())
        image.setColorTable(table);

    image.save(QString("data-%1.jpg").arg(id++), 0, 90);

    mem.close();
    mem.setData(QByteArray());

    } // while

    qDebug("shutting down");

    // canel operations
    sane_cancel(handle);
    sane_close(handle);

    // general shutdown
    sane_exit();

    qDebug("exit");

    return 0;
}
